package junctionbox;

import com.illposed.osc.OSCBundle;
import com.illposed.osc.OSCMessage;
import com.illposed.osc.OSCPortOut;
import java.net.InetAddress;
import processing.core.PConstants;

/**
 * The Simulator class produces TUIO cursor messages based on the location of the mouse cursor in Processing. Since the TUIO
 * messages are based on mouse location, only one touch is simulated.
 * 
 * @author Lawrence Fyfe
 * @version 0.98
 * @see <a href="http://tuio.org/?specification">TUIO 1.1 Protocol Specification</a>
 */
public class Simulator implements PConstants {
    private OSCPortOut out;
    private float mouseX;
    private float mouseY;
    private float boxWidth;
    private float boxHeight;
    private int sessionID = -1;
    private int frame = -1;
    private boolean set = false;
    private boolean alive = false;
    private float velocityX = 0.0f;
    private float velocityY = 0.0f;
    private float acceleration = 0.0f;
    private boolean newCursor = true;
    private long lastTime = 0;
    private long currentTime = 0;
    private float speed = 0.0f;
    private float lastSpeed = 0.0f;
    
    /**
     * Constructs a new Simulator. An OSC message sender is created that sends to localhost on port 3333.
     * 
     * @param w the box width
     * @param h the box height
     */
    public Simulator(float w, float h) {
        boxWidth = w;
        boxHeight = h;
        
        try {
            out = new OSCPortOut(InetAddress.getLocalHost(), 3333);
        }
        catch (Exception e) {
            e.printStackTrace();
        }
    }
    
    /**
     * Takes in data about the location of a mouse cursor.
     * 
     * @param pressed true if the mouse button is currently pressed
     * @param button the mouse button, LEFT, RIGHT or CENTER
     * @param x the x coordinate of the current mouse location
     * @param y the y coordinate of the currnet mouse location
     * @param px the x coordinate of the previous mouse location
     * @param py the y coordinate of the previous mouse location
     */
    public void mouse(boolean pressed, int button, float x, float y, float px, float py) {
        mouseX = x;
        mouseY = y;
        
        currentTime = System.currentTimeMillis();
        
        float dt = currentTime - lastTime;
        float dx = mouseX - px;
        float dy = mouseY - py;
        float d = (float)Math.sqrt(dx*dx + dy*dy);
        
        speed = d/dt;
        velocityX = dx/dt;
        velocityY = dy/dt;
        acceleration = (speed - lastSpeed)/dt;
        
        if (pressed) {
            alive = true;
            
            switch (button) {
                case LEFT:
                    if (newCursor) {
                        sessionID++;
                        set = true;
                        newCursor = false;
                    }
                    
                    if (mouseX != px || mouseY != py) {
                        set = true;
                    }
                    break;
                case RIGHT:
                    // Coming soon...
                    break;
                case CENTER:
                    // Not all systems have a third mouse button, so do nothing for now
                    break;
            }
        }
        else {
            alive = false;
            set = false;
            newCursor = true;
        }
        
        lastTime = currentTime;
        lastSpeed = speed;
        
        // Send messages
        send();
    }
    
    /**
     * Sends TUIO messages based on data about mouse location.
     */
    private void send() {
        OSCBundle bundle = new OSCBundle();
        
        // Alive message
        OSCMessage aliveMessage = new OSCMessage();
        aliveMessage.setAddress("/tuio/2Dcur");
        aliveMessage.addArgument("alive");
        
        // Only set this is the mouse button is down
        if (alive) {
            aliveMessage.addArgument(sessionID);
        }
        
        bundle.addPacket(aliveMessage);
        
        if (set) {
            OSCMessage setMessage = new OSCMessage();
            setMessage.setAddress("/tuio/2Dcur");
            setMessage.addArgument("set");
            setMessage.addArgument(sessionID);
            setMessage.addArgument(1 - ((boxWidth - mouseX)/boxWidth));
            setMessage.addArgument(1 - ((boxHeight - mouseY)/boxHeight));
            setMessage.addArgument(velocityX);
            setMessage.addArgument(velocityY);
            setMessage.addArgument(acceleration);
            
            bundle.addPacket(setMessage);
        }
        // Increase fseq for each new bundle
        frame++;
        
        // Each bundle has an fseq message
        OSCMessage fseqMessage = new OSCMessage();
        fseqMessage.setAddress("/tuio/2Dcur");
        fseqMessage.addArgument("fseq");
        fseqMessage.addArgument(frame);
        
        bundle.addPacket(fseqMessage);
        
        try {
            out.send(bundle);
        }
        catch (Exception e) {
            e.printStackTrace();
        }
    }
}
